/*
 * DatabaseImpl.java   E.L. 2005-09-24
 *
 * Klassene p denne filen gir tilgang til en database via en databasepool.
 * Databasepoolen administreres av et objekt av klassen DbWrapperFabrikkImpl.
 * Dette objektet er registrert i rmi-registeret. En klient henvender seg til
 * det objektet for  f sitt eget, eksklusive, databasewrapper-objekt. Ogs
 * det er et fjernobjekt.Alle databasewrapper-objektene bruker den samme databasepoolen.
 *
 * Feilhndtering: Alle metodene kan kaste RemoteException, som m behandles
 * p klientsiden. Metodene sjekker ikke returverdiene fra databasepool-metodene
 * for reservering og frigiving av databaseforbindelser.
 */

package mittBibliotek.database1_4;
import java.util.ArrayList;
import java.sql.*;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;


/*
 * Et objekt av klassen gir en forbindelse til en persondatabase.
 */
public class DatabaseImpl extends UnicastRemoteObject implements Database {
  private DatabasePool dbPool;
  private int kode; // Returkode fra databasesystemet dersom konflikt mellom primrnkkelverdier.

  public DatabaseImpl(DatabasePool startDbPool, int startKode) throws RemoteException {
    dbPool = startDbPool;
    kode = startKode;
  }

  /*
   * Returnerer en tabell-liste som inneholder alle dataene i tabellen person.
   * Tabell-listen inneholder personobjekter. Listen er sortert etter etternavn.
   * Returnerer null dersom SQLException er kastet. Da skrives ogs melding til konsollet.
   */
  public ArrayList finnAlle() throws RemoteException {
    ArrayList alle = new ArrayList();
    /* initcap() er en sql-funksjon som setter frste bokstav i hvert ord stor, resten smtt */
    String sqlsetning = "select persnr, initcap(fornavn) as fn, initcap(etternavn) as en " +
                        "from person order by etternavn, fornavn";
    System.out.println(sqlsetning);

    Forbindelse forbindelse = null;
    ResultSet res = null;
    Statement setning = null;
    try {
      forbindelse = dbPool.reserverForbindelse();
      setning = forbindelse.finnForbindelse().createStatement();
      res = setning.executeQuery(sqlsetning);
      while (res.next()) {
        int nr = res.getInt("persnr");
        String fornavn = res.getString("fn");
        String etternavn = res.getString("en");
        Person navnet = new Person(nr, fornavn, etternavn);
        alle.add(navnet);
      }
    } catch (Exception e) {
      skrivMelding(e, "finnAlle()");
      alle = null;
    } finally {  // setningene nedenfor utfres alltid
      lukkResSet(res);
      lukkSetning(setning);
      dbPool.frigiForbindelse(forbindelse.finnNr());
    }
    return alle;
  }

  /*
   * Endrer navn. Det nye navnet samt personens nummer sendes inn som argument.
   * Returnerer false dersom ingen person med dette nummeret eksisterer,
   * eller SQLException er inntruffet. I siste tilfelle skrives ogs melding til konsollet.
   */
  public boolean endreNavn(Person personen) throws RemoteException {
    int nr = personen.finnPersonNr();
    String nyttFornavn = personen.finnFornavn();
    String nyttEtternavn = personen.finnEtternavn();
    String sqlsetning = "update person set fornavn = upper('" + nyttFornavn
          + "'), etternavn = upper('" + nyttEtternavn + "') where persnr = " + nr;
    System.out.println(sqlsetning);

    Forbindelse forbindelse = null;
    Statement setning = null;
    try {
      forbindelse = dbPool.reserverForbindelse();
      setning = (forbindelse.finnForbindelse()).createStatement();
      if (setning.executeUpdate(sqlsetning) == 0) return false;
      else return true;
    } catch (Exception e) {
      skrivMelding(e, "endreNavn()");
    } finally {
      lukkSetning(setning);  // utfres selv om det er return-setninger rett over
      dbPool.frigiForbindelse(forbindelse.finnNr());
    }
    return false;
  }

  /*
   * Registrerer ny person.
   * Nummer tildeles av systemet, og et fullstendig personobjekt returneres til klienten.
   * Dersom tabellen er tom, settes personnr lik 1, ellers settes det lik
   * hittil strste nummer + 1.(P grunn av at data kan slettes,
   * kan denne metoden fre til hull i nummersekvensen.)
   */
  public Person registrerNyPerson(String fornavn, String etternavn) throws RemoteException {
    int nyttNr = 1; // dersom data ikke er i databasen
    boolean ok = true;
    do {
      Forbindelse forbindelse = null;
      ResultSet res = null;
      Statement setning = null;
      try {
        String sqlsetning = "select max(persnr) as maks from person";

        forbindelse = dbPool.reserverForbindelse();
        setning = (forbindelse.finnForbindelse()).createStatement();
        res = setning.executeQuery(sqlsetning);
         /*Hvis det ikke er data i tabellen, vil maks() returnere SQL NULL. getInt() vil omforme denne til 0,
         * og nyttNr blir dermed 1 dersom det ikke er data i databasen. Det er ok.  */
        res.next();
        nyttNr = res.getInt("maks") + 1;
        sqlsetning = "insert into person values(" + nyttNr + ", upper('"
              + fornavn + "'), upper('" + etternavn + "'))";
        System.out.println(sqlsetning);
        setning.executeUpdate(sqlsetning);
      } catch (SQLException e) {
        /* Dersom feilen skyldes at person med samme nummer eksisterer fra fr,
           m det bety at en annen klient har vrt inne i databasen mellom
           vre to sql-setninger. Vi kjrer derfor gjennom en gang til.
           (Dette er noe som vil inntreffe svrt sjelden.) */
        if (e.getErrorCode() == kode) ok = false;
        else {
          skrivMelding(e, "registrerNyPerson()");
          return null;   // RETUR, feil oppsttt (finally-blokken blir utfrt)
        }
      } finally {
        lukkResSet(res);
        lukkSetning(setning);
        dbPool.frigiForbindelse(forbindelse.finnNr());
      }
    } while (!ok);
    return new Person(nyttNr, fornavn, etternavn);
  }

  /*
   * Sletter person med gitt nummer.
   * Returnerer false dersom personen ikke finnes, eller SQLException inntruffet.
   * I siste tilfelle skrives melding til konsollet.
   */
  public boolean slettPerson(int nr) throws RemoteException {
    String sqlsetning = "delete from person where persnr = " + nr;
    System.out.println(sqlsetning);

    Forbindelse forbindelse = null;
    Statement setning = null;
    try {
      forbindelse = dbPool.reserverForbindelse();
      setning = (forbindelse.finnForbindelse()).createStatement();
      if (setning.executeUpdate(sqlsetning) == 0) return false;
      else return true;
    } catch (Exception e) {
      skrivMelding(e, "slettPerson()");
    } finally {
      lukkSetning(setning);
      dbPool.frigiForbindelse(forbindelse.finnNr());
    }
    return false;
  }

  private void lukkResSet(ResultSet res) {
    try {
      if (res != null) res.close();
    } catch (SQLException e) {
      skrivMelding(e, "lukkResSet()");
    }
  }

  private void lukkSetning(Statement stm) {
    try {
      if (stm != null) stm.close();
    } catch (SQLException e) {
      skrivMelding(e, "lukkSetning()");
    }
  }

  private void skrivMelding(Exception e, String metode) {
    System.err.println("*** Feil oppsttt i klassen Database, metode " + metode + ". ***");
    e.printStackTrace(System.err);
  }
}