/*
 * DatabasePool.java   E.L. 2004-03-02
 *
 * En databasepool bestr av en mengde forbindelser. Databaseparametre
 * (driver, navn, osv.) er argumenter til konstruktren. Poolen opprettes
 * med en bestemt kapasitet. Dersom en metode etterspr en forbindelse,
 * og det ikke er noen ledig i poolen, blir det laget en ny forbindelse.
 * Denne blir lukket nr den aktuelle metoden leverer forbindelsen tilbake.
 * Ellers m metoden lukkAlleForbindelser() kalles for  stenge poolen.
 *
 * Feilhndtering: Konstruktren kaster unntak dersom poolen ikke kan pnes.
 * Ellers hndteres unntak ved utskrift til konsollet.
 * Slike unntak skyldes programmeringsfeil og sendes ikke til klienten.
 */

package mittBibliotek.database1_4;
import java.util.ArrayList;
import java.sql.*;

public class DatabasePool {
  private ArrayList pool = new ArrayList();
  private int kapasitet;  // forbindelsene nummereres 0, 1, 2, .. ., (kapasitet - 1)
  private String dbNavn;
  private String brukernavn;   // alle forbindelsene bruker samme brukernavn
  private String passord;

  public DatabasePool(int startKap, String dbDriver, String startDbNavn,
                      String startBrukernavn, String startPassord) throws Exception {
    kapasitet = startKap;
    dbNavn = startDbNavn;
    brukernavn = startBrukernavn;
    passord = startPassord;
    try {
      Class.forName(dbDriver);  // laster driverklassene
      for (int nr = 0; nr < kapasitet; nr++) {
        pool.add(new Forbindelse(nr, dbNavn, brukernavn, passord));
      }
    } catch (Exception e) {
      skrivMelding(e, "konstruktr");
      lukkAlleForbindelser(); // lukker alle som hittil er pnet
      throw e;
    }
  }

  /*
   * Reserverer frste ledige forbindelse. Hvis ingen ledige, opprettes en ny.
   * Returnerer null hvis problemer.
   */
  public synchronized Forbindelse reserverForbindelse() {
    for (int i = 0; i < pool.size(); i++) {
      Forbindelse forb = (Forbindelse) pool.get(i);
      if (forb.erLedig()) {
        System.out.println("Reserverer forbindelse " + forb.finnNr());
        forb.settLedig(false);
        return forb;
      }
    }
    Forbindelse ny = null;  // ingen ledige, lager ny forbindelse
    try {
      Forbindelse forb = (Forbindelse) pool.get(pool.size() - 1);
      ny = new Forbindelse(forb.finnNr() + 1, dbNavn, brukernavn, passord);
      ny.settLedig(false);
      pool.add(ny);
      System.out.println("Reserverer forbindelse " + ny.finnNr());
    } catch (SQLException e) {
      skrivMelding(e, "reserverForbindelse()");
    }
    return ny;
  }

  /*
   * Frigir forbindelse. Hvis forbindelsen er opprettet i metoden over, blir den lukket,
   * og det tilhrende Forbindelse-objektet blir fjernet.
   * Metoden returner false dersom ugyldig nr.
   */
  public synchronized boolean frigiForbindelse(int nr) {
    System.out.println("Frigir forbindelse " + nr + ".");
    if (nr < kapasitet) {
      Forbindelse forb = (Forbindelse) pool.get(nr);
      forb.settLedig(true);
      return true;
    } else {
      for (int i = kapasitet; i < pool.size(); i++) {
        Forbindelse forb = (Forbindelse) pool.get(i);
        if (forb.finnNr() == nr) {
          try {
            forb.lukkForbindelse();
            pool.remove(i);
            System.out.println("Lukker forbindelse nr. " + nr + ".");
            return true;
          } catch (SQLException e) {
            skrivMelding(e, "frigiForbindelse()");
          }
        }
      }
      return false; // ugyldig nr
    }
  }

  public synchronized String lagUtskrift() {
    String resultat = "";
    for (int i = 0; i < pool.size(); i++) {
      resultat += pool.get(i).toString() + "\n";
    }
    return resultat;
  }

  public synchronized void lukkAlleForbindelser() {
    for (int i = 0; i < pool.size(); i++) {
      Forbindelse forb = (Forbindelse) pool.get(i);
      try {
        forb.lukkForbindelse();
        System.out.println("Lukker forbindelse nr. " + forb.finnNr() + ".");
      } catch (Exception e) {
        skrivMelding(e, "lukkAlleForbindelser()");
      }
    }
    pool.clear();
  }

  private void skrivMelding(Exception e, String metode) {
    System.err.println("*** Feil i klassen DatabasePool, metode " + metode + " ***");
    e.printStackTrace(System.err);
  }
}